/*
*   @package        : rlib
*   @module         : rcore
*   @author         : Richard [http://steamcommunity.com/profiles/76561198135875727]
*   @copyright      : (c) 2018 - 2020
*   @since          : 1.0.0
*   @website        : https://rlib.io
*   @docs           : https://docs.rlib.io
*   @file           : cl_init.lua
*
*   MIT License
*
*   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
*   LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
*   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
*   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
*   standard tables and localization
*/

local rlib                  = rlib
local base                  = rcore
local pf                    = base.manifest.prefix

/*
*   Localized rlib routes
*/

local helper                = rlib.h
local access                = rlib.a
local design                = rlib.d

/*
*   Localized translation func
*/

local function lang( ... )
    return rlib:lang( ... )
end

/*
    rnet  > sms  > umsg

    sends a message directly to the ply chat
*/

local function rn_sms_umsg( len, pl )
    local msg = net.ReadTable( )
    if not msg then return end
    chat.AddText( unpack( msg ) )
end
net.Receive( 'rlib.sms.umsg', rn_sms_umsg )

/*
    rnet  > sms  > notify

    sends a standard notification directly to the ply screen
*/

local function rn_sms_notify( len, pl )
    local args      = net.ReadTable( )
    local msg       = args and args[ 1 ] or false
    local cat       = args and args[ 2 ] or 1
    local pos       = args and args[ 3 ] or 1

    design:notify( msg, cat, pos )
end
net.Receive( 'rlib.sms.notify', rn_sms_notify )

/*
    rnet  > inform  > bc

    sends a slider notification directly to the ply screen
    slides in from the right side.
*/

local function rn_sms_inform( len, pl )
    local args      = net.ReadTable( )
    local cat       = args and args[ 1 ] or 1
    local msg       = args and args[ 2 ] or ''
    local title     = args and args[ 3 ] or lang( 'notify_title_def' )
    local dur       = args and isnumber( args[ 4 ] ) and args[ 4 ] or 5

    design:inform( cat, msg, title, dur )
end
net.Receive( 'rlib.sms.inform', rn_sms_inform )

/*
    rnet  > bubble  > bc

    sends a bubble msg that displays to the bottom right of the
    players screen
*/

local function rn_sms_bubble( len, pl )
    local args      = net.ReadTable( )
    local msg       = args and args[ 1 ] or ''
    local dur       = args and isnumber( args[ 2 ] ) and args[ 2 ] or nil
    local clr_box   = args and IsColor( args[ 3 ] ) and args[ 3 ] or nil
    local clr_txt   = args and IsColor( args[ 4 ] ) and args[ 4 ] or nil

    design:bubble( msg, dur, clr_box, clr_txt )
end
net.Receive( 'rlib.sms.bubble', rn_sms_bubble )

/*
    rnet > notification > rbubble

    sends a bubble msg that displays to the bottom right of the
    players screen
*/

local function rn_sms_rbubble( len, pl )
    local args      = net.ReadTable( )
    local data      = args and args[ 1 ] or ''

    design:rbubble( data )
end
net.Receive( 'rlib.sms.rbubble', rn_sms_rbubble )

/*
    rnet > notification > push

    notification supports icons; slides in from middle right.
    previously known as 'notifcator'
*/

local function rn_sms_push( len, pl )
    local args      = net.ReadTable( )
    local msg       = args and args[ 1 ] or false
    local title     = args and args[ 2 ] or false
    local ico       = args and args[ 3 ] or '*'

    design:push( msg, title, ico )
end
net.Receive( 'rlib.sms.push', rn_sms_push )

/*
    rnet > notification > sos
*/

local function rn_sms_sos( len, pl )
    local args      = net.ReadTable( )
    local msg       = args and args[ 1 ] or false
    local title     = args and args[ 2 ] or false
    local ico       = args and args[ 3 ] or '*'

    design:sos( msg, title, ico )
end
net.Receive( 'rlib.sms.sos', rn_sms_sos )

/*
    rnet > notification > nms
*/

local function rn_sms_nms( len, pl )
    local args      = net.ReadTable( )
    local msg       = args and args[ 1 ] or false
    local title     = args and args[ 2 ] or false
    local ico       = args and args[ 3 ] or '*'

    design:nms( msg, title, ico )
end
net.Receive( 'rlib.sms.nms', rn_sms_nms )

/*
*   get material data
*
*   returns material data based on the values provided
*   providing no second arg ( mod ) will make it return m_src
*
*   providing a mod returns m_mod.id_src
*
*   @ex     : rlib.m:get( 'btn_menu_steam', mod )
*           : rlib.m:get( 'btn_menu_steam', 'mod_str' )
*
*   @param  : str src
*   @param  : tbl, str mod
*   @return : tbl, str
*/

function base:mats_get( src, mod )
    if not isstring( src ) then return end
    local suffix = ''
    if mod then
        if isstring( mod ) and self.modules[ mod ] and self.modules[ mod ].id then
            suffix = self.modules[ mod ].id
        elseif istable( mod ) and mod.id then
            suffix = mod.id
        end
    end

    if isstring( mod ) and not suffix or suffix == '' then
        suffix = mod
    end

    suffix = suffix and suffix:lower( )

    local mat_ref = ( not suffix and 'm_' .. src ) or ( 'm_' .. suffix .. '_' .. src )

    return rlib.m[ mat_ref ] or mat_ref or ''
end

/*
*   get material index
*
*   returns material data based on the values provided
*   no second arg ( mod ) will make it return m_src
*
*   providing a mod returns m_mod.id_src
*
*   @ex     : rlib.m:get( 'btn_menu_steam', mod )
*           : rlib.m:get( 'btn_menu_steam', 'mod_str' )
*
*   @param  : tbl, str modsrc
*   @param  : bool bString
*   @return : tbl, str
*/

function base:mats_index( modsrc, bPath )
    local suffix = ''
    if modsrc then
        if isstring( modsrc ) and self.modules[ modsrc ] and self.modules[ modsrc ].id then
            suffix = self.modules[ modsrc ].id
        elseif istable( modsrc ) and modsrc.id then
            suffix = modsrc.id
        end
    end

    if isstring( modsrc ) and not suffix or suffix == '' then
        suffix = modsrc
    end

    suffix = suffix and suffix:lower( )

    if not modsrc then
        return rlib.m
    end

    local mat_ref = 'm_' .. suffix .. '_'

    local resp, cnt = { }, 0
    for k, v in pairs( rlib.m ) do
        if ( string.find( k, mat_ref, 1, true ) ~= nil ) then
            local id    = k
            id          = id:gsub( mat_ref, '' )
            resp[ id ]  = bPath and k.path or v
            cnt         = cnt + 1
        end
    end

    return resp, cnt

end

/*
*   module > register > permissions
*
*   register permissions for each module
*
*   @param  : tbl source
*/

local function module_register_perms( source )
    if source and not istable( source ) then
        local trcback = debug.traceback( )
        rlib:log( 2, 'cannot register permissions for modules, bad table\n%s', trcback )
        return
    end

    source = source or base.modules

    for v in helper.get.data( source ) do
        if not v.enabled or not v.permissions then continue end
        access:initialize( v.permissions )
    end
end
rhook.new.gmod( 'PostGamemodeLoaded', 'rcore_modules_perms_register', module_register_perms )